#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Replays an existing analysis onto the sheet
"""

from copy import copy


from mitosheet.utils import get_new_step_id
from mitosheet.steps.filter import execute_filter_column
from mitosheet.save_utils import read_and_upgrade_analysis
from mitosheet.steps import STEP_TYPE_TO_STEP
from mitosheet.errors import make_execution_error


REPLAY_ANALYSIS_UPDATE_EVENT = 'replay_analysis_update'
REPLAY_ANALYSIS_UPDATE_PARAMS = [
    'analysis_name',
    'import_summaries',
    'clear_existing_analysis',
]

def execute_replay_analysis_update(
        wsc,
        analysis_name,
        import_summaries,
        clear_existing_analysis
    ):
    """
    This function reapplies all the steps summarized in the passed step summaries, 
    which come from a saved analysis. 

    If any of the step summaries fails, this function tries to roll back to before
    it applied any of the stems

    If clear_existing_analysis is set to true, then this will clear the entire widget
    state container (except the initalize step) before applying the saved analysis.
    """ 
    # We only keep the intialize step only
    if clear_existing_analysis:
        wsc.steps = wsc.steps[:1]

    # If we're getting an event telling us to update, we read in the steps from the file
    analysis = read_and_upgrade_analysis(analysis_name)

    # When replaying an analysis with import events, you can also send over
    # new params to the import events to replace them. We replace them in the steps here
    if import_summaries is not None:
        for step_idx, params in import_summaries.items():
            for key, value in params.items():
                analysis['steps'][step_idx][key] = value  

    # We make a shallow copy of the steps, as none of the objects
    # will be changed by the step summaries we apply   
    old_steps = copy(wsc.steps)
    
    try:
        for _, step_summary in analysis['steps'].items():

            curr_step = wsc.steps[-1]

            step_type = step_summary['step_type']

            step_obj = STEP_TYPE_TO_STEP[step_type]

            # Get the params for this event
            params = {key: value for key, value in step_summary.items() if key in step_obj['params']}

            # If it's filter, we need to do a lot of extra work
            # so we sent it into the execute function directly
            if step_obj['step_type'] == 'filter_column':
                execute_filter_column(
                    wsc,
                    **params
                )
                wsc.curr_step_idx = len(wsc.steps) - 1
                # We always make a new step, so we also save the id of this step
                wsc.curr_step['step_id'] = get_new_step_id()
                continue 


            # Actually execute this event
            new_step = step_obj['execute'](curr_step, **params)
            
            # Save the params for this event
            for key, value in params.items():
                new_step[key] = value

            # Every step also needs an id, which we add
            # TODO: in the future, we should functionalize (and unify) the running of
            # steps, so we don't have to remember to do this everywhere
            new_step['step_id'] = get_new_step_id()

            wsc.steps.append(new_step)
            wsc.curr_step_idx = len(wsc.steps) - 1

    except Exception as e:
        print(e)
        # We remove all applied steps if there was an error
        wsc.steps = old_steps
        wsc.curr_step_idx = len(old_steps) - 1

        # And report a generic error to the user
        raise make_execution_error()


REPLAY_ANALYSIS_UPDATE = {
    'event_type': REPLAY_ANALYSIS_UPDATE_EVENT,
    'params': REPLAY_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_replay_analysis_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains functions for upgrading analyses!

NOTE: when adding a specific function for upgrading one version of a step to the other, the
format of the function should be:

def upgrade_<old_step_type>_<old_step_version>_to_<new_step_type>_<new_step_version>(step):
    ....
"""

from mitosheet._version import __version__


def upgrade_group_1_to_pivot_2(step):
    """
    Upgrades from a group 1 step to a pivot 2 step, simply
    by changing the names of the params.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "group", 
        "sheet_index": x, 
        "group_rows": [...], 
        "group_columns": [...], 
        "values": {...: ...}}}
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "pivot", 
        "sheet_index": old['sheet_index'], 
        "pivot_rows": old['group_rows'], 
        "pivot_columns": old['group_columns'], 
        "values": old['values']
    }
    """
    return {
        'step_version': 2,
        'step_type': 'pivot',
        'sheet_index': step['sheet_index'],
        'pivot_rows': step['group_rows'],
        'pivot_columns': step['group_columns'],
        'values': step['values']
    }

def upgrade_add_column_1_to_add_column_2(step): 
    """
    Upgrades from a add column 1 step to a add column 2 step, simply
    by adding the column_header_index param

    We just set the column_header_index to -1, so that it gets added to the 
    end of the dataframe, just like all previous analyses expect.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header'
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header', 
        'column_header_index'
    }
    """
    return {
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index': step['sheet_index'],
        'column_header': step['column_header'],
        'column_header_index': -1 # we set the column_header_index to -1 so that it gets added to the end
    }



"""
STEP_UPGRADES_FUNCTION_MAPPING mapping contains a mapping of all steps that need to be upgraded. A step
x at version y needs to be upgraded if STEP_UPGRADES[x][y] is defined, and in fact 
this mapping contains the function that can be used to do the upgrade!
 
NOTE: upgrades of steps should form a linear graph of upgrades to the most up to date
version. For example, if we change add_column from version 1 to version 2 to version 3, 
this object should contain:
    {
        'add_column': {
            1: upgrade_add_column_1_to_add_column_2,
            2: upgrade_add_column_2_to_add_column_3
        }
    }
"""
STEP_UPGRADES_FUNCTION_MAPPING = {
    'group': {
        1: upgrade_group_1_to_pivot_2
    }, 
    'add_column': {
        1: upgrade_add_column_1_to_add_column_2
    }
}


def is_prev_version(version: str, curr_version: str=__version__):
    """
    Returns True if the passed version is a previous version compared
    to the current version; note that this assumes semantic versioning
    with x.y.z!
    """
    old_version_parts = version.split('.')
    curr_version_parts = curr_version.split('.')

    for old_version_part, curr_version_part in zip(old_version_parts, curr_version_parts):
        if int(old_version_part) > int(curr_version_part):
            # E.g. if we have 0.2.11 and 0.1.11, we want to return early as it's clearly not older!
            return False

        if int(old_version_part) < int(curr_version_part):
            return True

    return False

def needs_upgrade(steps):
    """
    Returns True if _any_ of the steps in this step dictionary
    needs to be upgraded!
    """
    for step_idx, step in steps.items():
        step_version = step['step_version']
        step_type = step['step_type']

        if step_type in STEP_UPGRADES_FUNCTION_MAPPING:
            step_upgrades = STEP_UPGRADES_FUNCTION_MAPPING[step_type]
            if step_version in step_upgrades:
                return True

    return False


def upgrade_saved_analysis_to_current_version(saved_analysis):
    """
    Upgrades a saved analysis to the current version!
    """
    if saved_analysis is None:
        return None

    version = saved_analysis['version']
    steps = saved_analysis['steps']

    if not is_prev_version(version) or not needs_upgrade(steps):
        # TODO: do we want to _change_ the version this was saved with? I think it doesn't
        # really matter, as it gets changed when it gets rewritten...
        return saved_analysis

    new_steps = []
    for step_idx, step in steps.items():
        step_version = step['step_version']
        step_type = step['step_type']
        new_step = step

        if step_type in STEP_UPGRADES_FUNCTION_MAPPING:
            step_upgrades = STEP_UPGRADES_FUNCTION_MAPPING[step_type]
            if step_version in step_upgrades:
                new_step = step_upgrades[step_version](step)

        new_steps.append(new_step)

    # Convert the new steps in the correct format
    new_steps_json = {
        str(i + 1): step for i, step in enumerate(new_steps)
    }

    return {
        'version': __version__,
        'steps': new_steps_json
    }


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains handlers for the Mito API
"""
from mitosheet.mito_analytics import log_event_processed
import os
from queue import Queue
from threading import Thread
from pathlib import Path

from mitosheet.api.get_column_dtype import get_column_dtype
from mitosheet.api.get_graph import get_graph
from mitosheet.api.get_saved_analysis_description import get_saved_analysis_description
from mitosheet.api.get_column_describe import get_column_describe
from mitosheet.steps.import_steps.raw_python_import import RAW_PYTHON_IMPORT_STEP_TYPE
from mitosheet.steps.import_steps.simple_import import SIMPLE_IMPORT_STEP_TYPE
from mitosheet.save_utils import read_analysis

# As the column summary statistics tab does three calls, we defaulted to this max
MAX_QUEUED_API_CALLS = 3


class API():
    """
    The API provides a wrapper around a thread that responds to API calls.

    Some notes:
    -   We allow at most MAX_QUEUED_API_CALLS API calls to be in the queue, which practically
        Stops a backlog of calls from building up.
    -   All API calls should only be reads. This stops us from having to worry 
        about most concurrency issues
    -   Note that printing inside of a thread does not work properly! Use sys.stdout.flush() after the print statement.
        See here: https://stackoverflow.com/questions/18234469/python-multithreaded-print-statements-delayed-until-all-threads-complete-executi
    """
    def __init__(self, wsc, send):
        self.api_queue = Queue(MAX_QUEUED_API_CALLS)
        # Note that we make the thread a daemon thread, which practically means that when
        # The process that starts this thread terminate, our API will terminate as well.
        self.thread = Thread(target=handle_api_event_thread, args=(self.api_queue, wsc, send), daemon=True)
        self.thread.start()

    def process_new_api_call(self, event):
        """
        We privilege new API calls over old calls, and evict the old ones 
        if the API queue is full.

        Because we are using a queue, only events that have not been started 
        being processed will get removed.
        """        
        if self.api_queue.full():
            self.api_queue.get()
        self.api_queue.put(event)


def handle_api_event_thread(queue, wsc, send):
    """
    This is the worker thread function, that actually is 
    responsible for handling at the API call events.

    It lives forever, and just handles events as it 
    receives them from the queue
    """
    while True:
        # Note that this blocks when there is nothing in the queue,
        # and waits till there is something there - so no infinite 
        # loop as it is waiting!
        event = queue.get()
        # We place the API handling inside of a try catch, 
        # because otherwise if an error is thrown, then the entire thread crashes, 
        # and then the API never works again
        try:
            handle_api_event(send, event, wsc)
        except:
            # Log in error if it occurs
            log_event_processed(event, wsc, failed=True)


def get_filenames_with_suffix(suffix):
    """
    Returns all the file names in the current folder that end with the given
    suffix, sorted from most-recently created to the oldest.
    """
    # We sort them by creation time, to get the most recent files, as the user
    # is more likely to want these
    filenames = sorted(Path('.').iterdir(), key=os.path.getmtime)
    filenames.reverse()
    return [str(filename) for filename in filenames if filename.suffix == suffix]


def handle_datafiles(send, event):
    """
    Handles a `datafiles` api call, and returns all the csv files
    in the current folder.
    """
    csv_files = get_filenames_with_suffix('.csv')
    # TODO: also get the XLSX files, when we can import them
    send({
        'event': 'api_response',
        'id': event['id'],
        'data': csv_files
    })

def handle_import_summary(send, event):
    """
    Handle import summary is a route that, given the name of an analysis, will
    return the parameters to import steps over the course of the analysis. 

    The data we return is in the format:
    {
        "1": {
            "file_names": ["file123.csv"]
        }, 
        "3": {
            "python_code": "import pandas as ...",
            "new_df_names": ["df1"]
        }
    }
    which is a mapping from raw import steps to the files that they import.
    """
    analysis_name = event['analysis_name']
    # NOTE: we don't upgrade, as this happens when you actually choose to replay an analysis
    analysis = read_analysis(analysis_name)

    imports_only = dict()
    if analysis is not None:
        for step_idx, step in analysis['steps'].items():
            if step['step_type'] == SIMPLE_IMPORT_STEP_TYPE:
                imports_only[step_idx] = dict()
                imports_only[step_idx]['step_type'] = SIMPLE_IMPORT_STEP_TYPE
                imports_only[step_idx]['file_names'] = step['file_names']
            elif step['step_type'] == RAW_PYTHON_IMPORT_STEP_TYPE:
                imports_only[step_idx] = dict()
                imports_only[step_idx]['step_type'] = RAW_PYTHON_IMPORT_STEP_TYPE
                imports_only[step_idx]['python_code'] = step['python_code']
                imports_only[step_idx]['new_df_names'] = step['new_df_names']

    send({
        'event': 'api_response',
        'id': event['id'],
        'data': imports_only
    })


def get_dataframe_as_csv(send, event, wsc):
    """
    Sends a dataframe as a CSV string
    """
    sheet_index = event['sheet_index']
    df = wsc.dfs[sheet_index]

    send({
        'event': 'api_response',
        'id': event['id'],
        'data': df.to_csv(index=False)
    })


def handle_api_event(send, event, wsc):
    """
    Handler for all API calls. Note that any response to the
    API must return the same ID that the incoming message contains,
    so that the frontend knows how to match the responses.
    """    

    # And then handle it
    if event['type'] == 'datafiles':
        handle_datafiles(send, event)
    elif event['type'] == 'import_summary':
        handle_import_summary(send, event)
    elif event['type'] == 'get_dataframe_as_csv':
        get_dataframe_as_csv(send, event, wsc)
    elif event['type'] == 'get_graph':
        get_graph(send, event, wsc)
    elif event['type'] == 'get_column_describe':
        get_column_describe(send, event, wsc)
    elif event['type'] == 'get_column_dtype':
        get_column_dtype(send, event, wsc)
    elif event['type'] == 'get_saved_analysis_description':
        get_saved_analysis_description(send, event, wsc)
    else:
        raise Exception(f'Event: {event} is not a valid API call')


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A rename_column step, which allows you to rename a column
in a dataframe.
"""
from mitosheet.parser import safe_replace
from mitosheet.utils import create_new_step, does_sheet_index_exist_within_step, is_valid_header
from mitosheet.steps.column_steps.set_column_formula import _update_column_formula_in_step

from mitosheet.errors import (
    make_invalid_column_headers_error,
    make_column_exists_error, make_no_sheet_error,
)

RENAME_COLUMN_DISPLAY_NAME = 'Renamed a Column'

RENAME_COLUMN_EVENT = 'rename_column_edit'
RENAME_COLUMN_STEP_TYPE = 'rename_column'

RENAME_COLUMN_PARAMS = [
    'sheet_index', # the sheet to delete the column from
    'old_column_header', # the column to rename
    'new_column_header', # the new name for the column
]

    
def execute_rename_column_step(
        curr_step,
        sheet_index: int, 
        old_column_header: str, 
        new_column_header: str,
    ):
    """
    Renames the column from df at sheet_index from old_column_header to new_column_header.

    The majority of the work in this function is to update all the references to this other 
    column in all other areas of the step!
    """

    # if the sheet doesn't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    if not is_valid_header(new_column_header):
        raise make_invalid_column_headers_error([new_column_header])

    if new_column_header in curr_step['dfs'][sheet_index].keys():
        raise make_column_exists_error(new_column_header)

    # Create and checkout a rename step
    new_step = create_new_step(curr_step, RENAME_COLUMN_STEP_TYPE)

    # Execute the rename
    new_step['dfs'][sheet_index].rename(columns={old_column_header: new_column_header}, inplace=True)

    # Then, we update the current step to be valid, namely by deleting the old column (wherever it is)
    # and replacing it with the new column. 
    sheet_column_metatype = new_step['column_metatype'][sheet_index]
    sheet_column_metatype[new_column_header] = sheet_column_metatype[old_column_header]

    sheet_column_type = new_step['column_type'][sheet_index]
    sheet_column_type[new_column_header] = sheet_column_type[old_column_header]

    sheet_column_spreadsheet_code = new_step['column_spreadsheet_code'][sheet_index]
    sheet_column_spreadsheet_code[new_column_header] = sheet_column_spreadsheet_code[old_column_header]

    sheet_column_python_code = new_step['column_python_code'][sheet_index]
    sheet_column_python_code[new_column_header] = sheet_column_python_code[old_column_header].replace(
        f'df[\'{old_column_header}\']',
        f'df[\'{new_column_header}\']'
    )
    
    sheet_column_evaluation_graph = new_step['column_evaluation_graph'][sheet_index]
    sheet_column_evaluation_graph[new_column_header] = sheet_column_evaluation_graph[old_column_header]

    sheet_column_filters = new_step['column_filters'][sheet_index]
    sheet_column_filters[new_column_header] = sheet_column_filters[old_column_header]

    # We also have to go over _all_ the formulas in the sheet that reference this column, and update
    # their references to the new column. 
    for column_header in sheet_column_evaluation_graph[new_column_header]:
        old_formula = sheet_column_spreadsheet_code[column_header]
        new_formula = safe_replace(
            old_formula,
            old_column_header,
            new_column_header
        )

        # NOTE: this does not update the evaluation graph for columns that are descendents
        # of this column, so we do that below.
        _update_column_formula_in_step(new_step, sheet_index, column_header, old_formula, new_formula)

    # We then have to go through and update the evaluation graphs
    # for the columns the renamed column relied on.
    for dependents in sheet_column_evaluation_graph.values():
        if old_column_header in dependents:
            dependents.remove(old_column_header)
            dependents.add(new_column_header)

    # We delete all references to the old_column header
    # NOTE: this has to happen after the above formula setting, so that
    # the dependencies can be updated properly!
    del sheet_column_metatype[old_column_header]
    del sheet_column_type[old_column_header]
    del sheet_column_spreadsheet_code[old_column_header]
    del sheet_column_python_code[old_column_header]
    del sheet_column_evaluation_graph[old_column_header]
    del sheet_column_filters[old_column_header]

    return new_step


def transpile_rename_column_step(
        step,
        sheet_index,
        old_column_header,
        new_column_header
    ):
    df_name = step["df_names"][sheet_index]
    rename_dict = "{\"" + old_column_header + "\": \"" + new_column_header + "\"}"
    return [f'{df_name}.rename(columns={rename_dict}, inplace=True)']


def describe_rename_column_step(
        sheet_index,
        old_column_header,
        new_column_header,
        df_names=None
    ):

    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Renamed {old_column_header} to {new_column_header} in {df_name}'
    return f'Renamed {old_column_header} to {new_column_header}'


RENAME_COLUMN_STEP = {
    'step_version': 1,
    'step_display_name': RENAME_COLUMN_DISPLAY_NAME,
    'event_type': RENAME_COLUMN_EVENT,
    'step_type': RENAME_COLUMN_STEP_TYPE,
    'params': RENAME_COLUMN_PARAMS,
    'saturate': None,
    'execute': execute_rename_column_step,
    'transpile': transpile_rename_column_step,
    'describe': describe_rename_column_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions for saving and reading
in analyses.
"""

from mitosheet.upgrade import upgrade_saved_analysis_to_current_version
import os
import json

from mitosheet._version import __version__
from mitosheet.steps import STEPS
from mitosheet.mito_analytics import log
from mitosheet.saves import SAVES


# Where all global .mito files are stored
MITO_FOLDER = os.path.expanduser("~/.mito")

# The current version of the saved Mito analysis
# where we save all the analyses for this version
SAVED_ANALYSIS_FOLDER = os.path.join(MITO_FOLDER, 'saved_analyses')


def read_analysis(analysis_name):
    """
    Given an analysis_name, reads the saved analysis in
    ~/.mito/{analysis_name}.json and returns a JSON object
    representing it.

    Also, checks if the analysis name is in the list of Mito
    saves, and will return that if so.
    """
    if analysis_name in SAVES:
        return SAVES[analysis_name]

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'

    if not os.path.exists(analysis_path):
        return None

    with open(analysis_path) as f:
        try:
            # We try and read the file as JSON
            return json.load(f)
        except: 
            return None

def read_and_upgrade_analysis(analysis_name):
    """
    Given an analysis_name, reads the saved analysis in
    ~/.mito/{analysis_name}.json, does it's best to upgrade it to the current
    saved version, and then returns it.
    """
    old_analysis = read_analysis(analysis_name)
    return upgrade_saved_analysis_to_current_version(old_analysis)

def _get_all_analysis_filenames():
    """
    Returns the names of the files in the SAVED_ANALYSIS_FOLDER
    """
    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        return []

    file_names = set([
        f for f in os.listdir(SAVED_ANALYSIS_FOLDER) 
        if os.path.isfile(os.path.join(SAVED_ANALYSIS_FOLDER, f))
    ])

    return file_names

def _delete_analyses(analysis_filenames):
    """
    For bulk deleting analysis with file names. 
    """
    for filename in analysis_filenames:
        os.remove(os.path.join(SAVED_ANALYSIS_FOLDER, filename))

def delete_saved_analysis(analysis_name):
    """
    Deletes a saved analysis. saved_analysis_file_name must end in .json

    Throws an error if analysis_name does not exist.
    """

    analysis = read_analysis(analysis_name)

    # If the saved_analysis_name exists, delete it
    if analysis is not None:
        os.remove(os.path.join(SAVED_ANALYSIS_FOLDER, analysis_name + '.json'))
    else:
        raise Exception(f'Cannot delete {analysis_name} as it does not exist')


def rename_saved_analysis(old_analysis_name, new_analysis_name):
    """
    Renames a saved analysis from old_analysis_name to new_analysis_name. 

    Throws an error if old_analysis_name does not exist, or new_analysis_name
    exists.
    """
    old_analysis = read_analysis(old_analysis_name)
    new_analysis = read_analysis(new_analysis_name)

    # If the old_analysis_file_name exists, and new_analysis_file_name does not exist
    if old_analysis is not None and new_analysis is None:
        full_old_analysis_name = os.path.join(SAVED_ANALYSIS_FOLDER, old_analysis_name + '.json')
        full_new_analysis_name = os.path.join(SAVED_ANALYSIS_FOLDER, new_analysis_name + '.json')
        os.rename(full_old_analysis_name, full_new_analysis_name)
    else:
        raise Exception(f'Invalid rename, with old and new analysis are {old_analysis_name} and {new_analysis_name}')


def read_saved_analysis_names():
    """
    Reads the names of all the analyses saved by the user.

    Does not return any of the auto-saved analyses!
    """
    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        return []

    file_names = [
        f for f in os.listdir(SAVED_ANALYSIS_FOLDER) 
        if os.path.isfile(os.path.join(SAVED_ANALYSIS_FOLDER, f))
        and not f.startswith('UUID-')
    ]

    # We make sure they are in alphabetical order!
    file_names.sort()

    return [
        file_name[:-5] for file_name in file_names 
        if file_name.endswith('.json')
    ]

def saved_analysis_names_json():
    return json.dumps(read_saved_analysis_names())

def make_steps_json_obj(steps):
    """
    Given a steps dictonary from a widget_state_container, puts the steps
    into a format that can be saved and recreated. Necessary for saving an
    analysis to a file!
    """
    steps_json_obj = dict()
    for step_idx, step in enumerate(steps):
        step_type = step['step_type']
        
        for new_step in STEPS:
            if step_type == new_step['step_type']:

                # Save the step type
                step_summary = {
                    'step_version': new_step['step_version'],
                    'step_type': step_type,
                }
                # As well as all of the parameters for the step
                step_summary.update({key: value for key, value in step.items() if key in new_step['params']})

                steps_json_obj[step_idx] = step_summary

    return steps_json_obj

def write_saved_analysis(analysis_path, steps, version=__version__):
    with open(analysis_path, 'w+') as f:
        saved_analysis = {
            'version': version,
            'steps': steps
        }

        f.write(json.dumps(saved_analysis))

def write_analysis(widget_state_container, analysis_name=None):
    """
    Writes the analysis saved in widget_state_container to
    ~/.mito/{analysis_name}. If analysis_name is none, gets the temporary
    name from the widget_state_container.

    NOTE: as the written analysis is from the widget_state_container,
    we assume that the analysis is valid when written and read back in!
    """

    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        os.mkdir(SAVED_ANALYSIS_FOLDER)

    if analysis_name is None:
        analysis_name = widget_state_container.analysis_name

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'
    steps = make_steps_json_obj(widget_state_container.steps)

    # Actually write the file
    write_saved_analysis(analysis_path, steps)


def register_analysis(analysis_name):
    """
    A helper function that is used in the frontend
    to save which analysis is being run (so it can be replayed)
    """
    log('ran_generated_code', {'analysis_name': analysis_name}) 


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A reorder_column step, which allows you to move 
a column to a different location in the df.
"""
from mitosheet.utils import create_new_step, does_sheet_index_exist_within_step
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
)

REORDER_COLUMN_DISPLAY_NAME = 'Reordered Columns'

REORDER_COLUMN_EVENT = 'reorder_column_edit'
REORDER_COLUMN_STEP_TYPE = 'reorder_column'

REORDER_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # the column to reorder
    'new_column_index', # the new location for the column
]

def get_valid_index(dfs, sheet_index, new_column_index):
    # make sure new_column_index is valid
    if new_column_index < 0:
        new_column_index = 0

    if new_column_index >= len(dfs[sheet_index].columns):
        new_column_index = len(dfs[sheet_index].columns) - 1

    return new_column_index

def execute_reorder_column_step(
        curr_step,
        sheet_index,
        column_header,
        new_column_index
    ):
    """
    The function responsible for updating the widget state container
    with a new column reorder step.

    If it fails part of the way through, deletes the new column reorder step entirely.
    """
    # if the sheet doesn't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the column to be reordered exists
    missing_column = set([column_header]).difference(curr_step['column_metatype'][sheet_index].keys())
    if len(missing_column) > 0:
        raise make_no_column_error(missing_column)

    new_column_index = get_valid_index(curr_step['dfs'], sheet_index, new_column_index)
        
    # Create a new step
    new_step = create_new_step(curr_step, REORDER_COLUMN_STEP_TYPE)

    # Actually execute the column reordering
    new_step['dfs'][sheet_index] = _execute_reorder_column(
        new_step['dfs'][sheet_index],
        column_header,
        new_column_index
    )

    return new_step


def _execute_reorder_column(df, column_header, new_column_index):
    """
    Helper function for reordering a column in the dataframe
    """
    df_columns = [col for col in df.columns if col != column_header]
    df_columns.insert(new_column_index, column_header)
    return df[df_columns]


def transpile_reorder_column_step(
        step,
        sheet_index,
        column_header,
        new_column_index
    ):
    """
    Transpiles a column reorder step to python code!
    """

    new_column_index = get_valid_index(step['dfs'], sheet_index, new_column_index)

    # get columns in df
    columns_list_line = f'{step["df_names"][sheet_index]}_columns = [col for col in {step["df_names"][sheet_index]}.columns if col != \'{column_header}\']'

    # insert column into correct location 
    insert_line = f'{step["df_names"][sheet_index]}_columns.insert({new_column_index}, \'{column_header}\')'
    
    # Apply reorder line
    apply_reorder_line = f'{step["df_names"][sheet_index]} = {step["df_names"][sheet_index]}[{step["df_names"][sheet_index]}_columns]'

    return [columns_list_line, insert_line, apply_reorder_line]


def describe_reorder_column_step(
        sheet_index,
        column_header,
        new_column_index,
        df_names=None
    ):

    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Reordered {column_header} in {df_name}'
    return f'Reordered {column_header}'


REORDER_COLUMN_STEP = {
    'step_version': 1,
    'step_display_name': REORDER_COLUMN_DISPLAY_NAME,
    'event_type': REORDER_COLUMN_EVENT,
    'step_type': REORDER_COLUMN_STEP_TYPE,
    'params': REORDER_COLUMN_PARAMS,
    'saturate': None,
    'execute': execute_reorder_column_step,
    'transpile': transpile_reorder_column_step,
    'describe': describe_reorder_column_step
}


from enum import Enum
import pandas as pd


"""
    The DataTypeInMito enum is used to signify the type of data in Mito.
    It can be one of four types:

    1. NONE = no data is in Mito
    2. PROVIDED = any data that we provided through our docs, except the tutorial data which has a special designation
    3. TUTORIAL = the tutorial data provided through the docs
    4. PERSONAL = any data not provided by Mito

    NOTE: this should be the same as the enum in Mito.tsx with the same name
"""
class DataTypeInMito(Enum):
    NONE = 'none'
    PROVIDED = 'provided'
    TUTORIAL = 'tutorial'
    PERSONAL = 'personal'


# We keep track of the tutorial and provided data
TUTORIAL_FILE_NAMES = ['Airport-Pets.csv', 'Zipcode-Data.csv']
TUTORIAL_AIRPORT_PETS_COLUMNS = ['Zip', 'City', 'State', 'Division', 'Parking', 'Pets', 'Food', 'Lounge']
TUTORIAL_ZIPCODE_COLUMNS = ['Zip', 'Median_Income', 'Mean_Income', 'Pop']

PROVIDED_TICKET_OFFICE_COLUMNS = ['Zip', 'City', 'State', 'Ticket_Office']
PROVIDED_TICKET_ZIPCODE_COLUMNS = ['Zip', 'Median_Income', 'Median_Income', 'Mean_Income', 'Pop']


def get_data_type_in_mito(*args) -> DataTypeInMito:
    """
    Returns the DataTypeInMito based on the args to the function.

    If this function is called by the simple import condition of the handle_edit_event,
    then the args are the names of the files that are imported

    If this function is called from the mitosheet.sheet() call then its either a string or dataframe object
    """
    args = list(args)
    tutorial_data_found = False

    if len(args) <= 0:
        return DataTypeInMito.NONE

    for arg in args:
        # If the arg is a string, its either the path to a file which we then assume is personal data or
        # it is the name of the file imported during a simple import. 
        if isinstance(arg, str):
            # If the arg is a string, check if its a tutorial file name.
            # If its not tutorial data, then its personal data
            if arg not in TUTORIAL_FILE_NAMES:
                return DataTypeInMito.PERSONAL
            else:
                # We mark that we came across tutorial data because if we don't end up finding personal data
                # then we want to be able to return that we found tutorial data
                tutorial_data_found = True

        # If the arg is a dataframe, then check if the columns are either from the tutorial or provided data. 
        if isinstance(arg, pd.DataFrame):
            # If the user passed a dataframe with headers the same as the tutorial data, then return DataTypeInMito.TUTORIAL
            if (
                arg.columns.tolist() == TUTORIAL_AIRPORT_PETS_COLUMNS or 
                arg.columns.tolist() == TUTORIAL_ZIPCODE_COLUMNS 
            ): 
                return DataTypeInMito.TUTORIAL

            # If the user passed a dataframe with headers the same as the provided data, then return DataTypeInMito.PROVIDED
            if (
                arg.columns.tolist() == PROVIDED_TICKET_OFFICE_COLUMNS or 
                arg.columns.tolist() == PROVIDED_TICKET_ZIPCODE_COLUMNS
            ):
                return DataTypeInMito.PROVIDED

    # If we came across tutorial data and have not yet returned Personal data, then return Tutorial Data
    # otherwise, the data is personal data
    return DataTypeInMito.TUTORIAL if tutorial_data_found else DataTypeInMito.PERSONAL



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains tests for the user.json file, making sure it upgrades properly,
and that it can handle the various undefined versions of user.json that exist.
"""
from datetime import datetime, timedelta
import pytest
import json
import os

from mitosheet._version import __version__
from mitosheet.user.user import initialize_user, should_upgrade_mitosheet
from mitosheet.user.user_utils import USER_JSON_PATH, get_user_field

@pytest.fixture(scope="module", autouse=True)
def cleanup_files():
    """
    This fixture reads in the original user.json file that exists before these tests are
    run, deletes it, and then recreates it at the end. This allows us to test what happens 
    when the user.json file is in various states of out of date and disrepair.
    """
    with open(USER_JSON_PATH, 'r') as f:
        user_json = json.loads(f.read())

    os.remove(USER_JSON_PATH)

    yield # All tests in this user module run right here

    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps(user_json))


def test_initialize_user_creates_json_when_empty():
    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert len(get_user_field('static_user_id')) > 0
    assert len(get_user_field('user_email')) == 0 or (get_user_field('user_email') == 'github@action.com' and 'CI' in os.environ)
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == [datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)


def test_initialize_upgrades_user_when_old_version():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'static_user_id': 'github_action',
            'user_email': 'github@action.com'
        }))

    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert get_user_field('static_user_id') == 'github_action'
    assert get_user_field('user_email') == 'github@action.com'
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == [datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)


def test_initalize_detects_new_usage_when_less_than_five():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': datetime.today().strftime('%Y-%m-%d'),
            'mitosheet_last_five_usages': ['2020-12-1']
        }))

    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert get_user_field('static_user_id') == 'github_action'
    assert get_user_field('user_email') == 'github@action.com'
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == ['2020-12-1', datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)


def test_initalize_detects_new_usage_when_five():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': datetime.today().strftime('%Y-%m-%d'),
            'mitosheet_last_five_usages': ['2020-12-1', '2020-12-2', '2020-12-3', '2020-12-4', '2020-12-5']
        }))

    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert get_user_field('static_user_id') == 'github_action'
    assert get_user_field('user_email') == 'github@action.com'
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == ['2020-12-2', '2020-12-3', '2020-12-4', '2020-12-5', datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)


def test_initialize_user_only_stores_day_once():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': datetime.today().strftime('%Y-%m-%d'),
            'mitosheet_last_five_usages': ['2020-12-1', '2020-12-2', '2020-12-3', '2020-12-4', datetime.today().strftime('%Y-%m-%d')]
        }))

    initialize_user()
    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert get_user_field('static_user_id') == 'github_action'
    assert get_user_field('user_email') == 'github@action.com'
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == ['2020-12-1', '2020-12-2', '2020-12-3', '2020-12-4', datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)

    
def test_initalize_detects_when_mito_upgraded():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': '0.1.100',
            'mitosheet_last_upgraded_date': '2020-12-1',
            'mitosheet_last_five_usages': ['2020-12-1', '2020-12-2', '2020-12-3', '2020-12-4', '2020-12-5']
        }))

    initialize_user()

    assert os.path.exists(USER_JSON_PATH)
    assert get_user_field('user_json_version') == 1
    assert get_user_field('static_user_id') == 'github_action'
    assert get_user_field('user_email') == 'github@action.com'
    assert get_user_field('mitosheet_current_version') == __version__
    assert get_user_field('mitosheet_last_upgraded_date') == datetime.today().strftime('%Y-%m-%d')
    assert get_user_field('mitosheet_last_five_usages') == ['2020-12-2', '2020-12-3', '2020-12-4', '2020-12-5', datetime.today().strftime('%Y-%m-%d')]

    os.remove(USER_JSON_PATH)


def test_should_not_upgrade_on_first_creation():
    initialize_user()
    assert not should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_not_upgrade_with_less_than_four_usages():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': '0.1.100',
            'mitosheet_last_upgraded_date': (datetime.today() - timedelta(days=5)).strftime('%Y-%m-%d'), # only 5 days ago,
            'mitosheet_last_five_usages': ['2020-12-1', '2020-12-2']
        }))

    initialize_user()
    assert not should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_upgrade_mitosheet_greater_than_two_weeks():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': '2020-12-2',
            'mitosheet_last_five_usages': ['2020-12-1', '2020-12-2']
        }))

    initialize_user()
    assert should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_upgrade_mitosheet_greater_than_two_weeks_and_five_usages_since_then():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': '2020-12-2',
            'mitosheet_last_five_usages': ['2020-12-3', '2020-12-4', '2020-12-5', '2020-12-6', '2020-12-7']
        }))

    initialize_user()
    assert should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_upgrade_mitosheet_greater_than_two_weeks_no_usage():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': '2020-12-2',
            'mitosheet_last_five_usages': []
        }))

    initialize_user()
    assert should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_upgrade_mitosheet_at_least_4_usages_since_last_upgrade():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': (datetime.today() - timedelta(days=5)).strftime('%Y-%m-%d'), # only 5 days ago
            'mitosheet_last_five_usages': [
                (datetime.today() - timedelta(days=4)).strftime('%Y-%m-%d'),
                (datetime.today() - timedelta(days=3)).strftime('%Y-%m-%d'),
                (datetime.today() - timedelta(days=2)).strftime('%Y-%m-%d')
            ]
        }))

    initialize_user()
    assert should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)


def test_should_not_upgrade_mitosheet_more_than_4_usages_but_some_before_last_upgrade():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'user_json_version': 1,
            'static_user_id': 'github_action',
            'user_email': 'github@action.com',
            'mitosheet_current_version': __version__,
            'mitosheet_last_upgraded_date': (datetime.today() - timedelta(days=5)).strftime('%Y-%m-%d'), # only 5 days ago
            'mitosheet_last_five_usages': [
                (datetime.today() - timedelta(days=6)).strftime('%Y-%m-%d'),
                (datetime.today() - timedelta(days=3)).strftime('%Y-%m-%d'),
                (datetime.today() - timedelta(days=2)).strftime('%Y-%m-%d')
            ]
        }))

    initialize_user()
    assert not should_upgrade_mitosheet()
    os.remove(USER_JSON_PATH)





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest
from mitosheet._version import __version__
from mitosheet.upgrade import (
    is_prev_version
)

from mitosheet.save_utils import (
    SAVED_ANALYSIS_FOLDER, write_saved_analysis,
    read_and_upgrade_analysis
)


PREV_TESTS = [
    ('0.1.61', '0.1.62', True),
    ('0.1.61', '0.1.610', True),
    ('0.1.62', '0.1.62', False),
    ('0.2.61', '0.1.62', False),
    ('0.2.62', '0.2.62', False),
]
@pytest.mark.parametrize("prev, curr, result", PREV_TESTS)
def test_prev_analysis_returns_correct_results(prev, curr, result):
    assert is_prev_version(prev, curr_version=curr) == result

TEST_ANALYSIS_NAME = 'UUID-test_analysis'
TEST_FILE = f'{SAVED_ANALYSIS_FOLDER}/{TEST_ANALYSIS_NAME}.json'

UPGRADE_TESTS = [
    # Simple column add step should update to add_column version 2
    (
        {
            "version": "0.1.59", "steps": {"1": {"step_version": 1, "step_type": "add_column", "sheet_index": 0, "column_header": "["}}
        },
        {
            "version": __version__, "steps": {"1": {"step_version": 2, "step_type": "add_column", "sheet_index": 0, "column_header": "[", "column_header_index": -1}}
        }
    ),
    # Add columns, set formulas, and delete
    (
        {
            "version": "0.1.55", 
            "steps": {"1": {"step_version": 1, "step_type": "add_column", "sheet_index": 0, "column_header": "D"}, "2": {"step_version": 1, "step_type": "add_column", "sheet_index": 0, "column_header": "E"}, "3": {"step_version": 1, "step_type": "set_column_formula", "sheet_index": 0, "column_header": "D", "old_formula": "=0", "new_formula": "=OFFSET(B, -1)"}, "4": {"step_version": 1, "step_type": "delete_column", "sheet_index": 0, "column_header": "D"}}
        },
        {
            "version": __version__, 
            "steps": {"1": {"step_version": 2, "step_type": "add_column", "sheet_index": 0, "column_header": "D", "column_header_index": -1}, "2": {"step_version": 2, "step_type": "add_column", "sheet_index": 0, "column_header": "E", "column_header_index": -1}, "3": {"step_version": 1, "step_type": "set_column_formula", "sheet_index": 0, "column_header": "D", "old_formula": "=0", "new_formula": "=OFFSET(B, -1)"}, "4": {"step_version": 1, "step_type": "delete_column", "sheet_index": 0, "column_header": "D"}}
        }
    ),
    # A merge should not be upgraded
    (
        {
            "version": "0.1.55", 
            "steps": {"1": {"step_version": 1, "step_type": "merge", "sheet_index_one": 0, "sheet_index_two": 1, "merge_key_one": "Name", "merge_key_two": "Name", "selected_columns_one": ["Name", "Number"], "selected_columns_two": ["Name", "Sign"]}, "2": {"step_version": 1, "step_type": "sort", "sheet_index": 2, "column_header": "Number", "sort_direction": "descending"}}
        },
        {
            "version": "0.1.55", 
            "steps": {"1": {"step_version": 1, "step_type": "merge", "sheet_index_one": 0, "sheet_index_two": 1, "merge_key_one": "Name", "merge_key_two": "Name", "selected_columns_one": ["Name", "Number"], "selected_columns_two": ["Name", "Sign"]}, "2": {"step_version": 1, "step_type": "sort", "sheet_index": 2, "column_header": "Number", "sort_direction": "descending"}}
        },
    ),
    # Test filter and sort
    (
        {
            "version": "0.1.55", 
            "steps": {"1": {"step_version": 1, "step_type": "filter_column", "sheet_index": 0, "column_header": "String", "filters": [{"type": "string", "condition": "contains", "value": "1"}], "operator": "And"}, "2": {"step_version": 1, "step_type": "sort", "sheet_index": 0, "column_header": "String", "sort_direction": "ascending"}}
        },
        {
            "version": "0.1.55", 
            "steps": {"1": {"step_version": 1, "step_type": "filter_column", "sheet_index": 0, "column_header": "String", "filters": [{"type": "string", "condition": "contains", "value": "1"}], "operator": "And"}, "2": {"step_version": 1, "step_type": "sort", "sheet_index": 0, "column_header": "String", "sort_direction": "ascending"}}
        },
    ),
    # Single group step, it should should be upgraded to a pivot
    (
        {
            "version": "0.1.54", 
            "steps": {"1": {"step_version": 1, "step_type": "group", "sheet_index": 0, "group_rows": ["Name"], "group_columns": [], "values": {"Height": "sum"}}}
        },
        {
            "version": __version__, 
            "steps": {"1": {"step_version": 2, "step_type": "pivot", "sheet_index": 0, "pivot_rows": ["Name"], "pivot_columns": [], "values": {"Height": "sum"}}}
        }
    ), 
    # Ends in a group step, should be upgraded to a pivot
    (
        {
            "version": "0.1.54", 
            "steps": {"1": {"step_version": 1, "step_type": "add_column", "sheet_index": 0, "column_header": "B"}, "2": {"step_version": 1, "step_type": "set_column_formula", "sheet_index": 0, "column_header": "B", "old_formula": "=0", "new_formula": "=1"}, "3": {"step_version": 1, "step_type": "group", "sheet_index": 0, "group_rows": ["Name"], "group_columns": ["DORK"], "values": {"Height": "sum"}}}
        },
        {
            "version": __version__, 
            "steps": {"1": {"step_version": 2, "step_type": "add_column", "sheet_index": 0, "column_header": "B", "column_header_index": -1}, "2": {"step_version": 1, "step_type": "set_column_formula", "sheet_index": 0, "column_header": "B", "old_formula": "=0", "new_formula": "=1"}, "3": {"step_version": 2, "step_type": "pivot", "sheet_index": 0, "pivot_rows": ["Name"], "pivot_columns": ["DORK"], "values": {"Height": "sum"}}}
        },
    ), 
    # A group step in the middle, should be upgraded to a pivot
    (
        {
            "version": "0.1.60", 
            "steps": {"1": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}, "2": {"step_version": 1, "step_type": "group", "sheet_index": 0, "group_rows": [], "group_columns": ["Ha", 'Ha'], "values": {}}, "3": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}}
        },
        {
            "version": __version__, 
            "steps": {"1": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}, "2": {"step_version": 2, "step_type": "pivot", "sheet_index": 0, "pivot_rows": [], "pivot_columns": ["Ha", "Ha"], "values": {}}, "3": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}}
        },
    ), 
    # An add_column step version 1 should be upgraded to add_column step version 2
    (
        {
            "version": "0.1.140", 
            "steps": {"1": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}, "2": {"step_version": 1, "step_type": "add_column", "sheet_index": 0, "column_header": 'New_Column_Header'}}
        },
        {
            "version": __version__, 
            "steps": {"1": {"step_version": 1, "step_type": "simple_import", "file_names": ["NamesNew.csv"]}, "2": {"step_version": 2, "step_type": "add_column", "sheet_index": 0, "column_header": 'New_Column_Header', "column_header_index": -1}}
        },
    ), 

]
@pytest.mark.parametrize("old, new", UPGRADE_TESTS)
def test_does_not_upgrade_current_analysis(old, new):
    write_saved_analysis(TEST_FILE, old['steps'], version=old['version'])
    assert read_and_upgrade_analysis(TEST_ANALYSIS_NAME) == new


# TODO: 
# 2. Add a test for _each_ of the step types!


import io
import base64
import pandas as pd
from mitosheet.sheet_functions.types.utils import get_mito_type
import matplotlib.pyplot as plt
import seaborn as sns
import sys # In order to print in this file use sys.stdout.flush() once after the print statements.

def get_graph(send, event, wsc):
    """
    Creates a graph of the passed parameters, and sends it back as a PNG
    string to the frontend for display.

    Requires params: column_header_x_axis, sheet_index_x_axis. Optionally takes the same
    params for column_header_y_axis and sheet_index_y_axis.

    If only an x axis is given, and if the series is a numeric series,
    will return a histogram. Otherwise, as long as there are less than 
    20 distinct items in the series, will return a bar chart of the 
    value count. Otherwise, will return nothing.
    """
    # We turn off interactive plotting for this function, to avoid graphs 
    # being printed to the console. However we make sure to turn it back on 
    # at the end if it was originally on, to not mess with any users code 
    # in the rest of their notebook
    was_interactive = plt.isinteractive()
    if was_interactive:
        plt.ioff()

    keys = event.keys()

    # If an x axis was not provided, then return an empty string
    if 'column_header_x_axis' not in keys or 'sheet_index_x_axis' not in keys:
        # Send an empty response
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': ''
        })
        # Then return so we only respond with an empty string
        return

    sheet_index_x_axis = event['sheet_index_x_axis']
    column_header_x_axis = event['column_header_x_axis']
    x_series: pd.Series = wsc.dfs[sheet_index_x_axis][column_header_x_axis]

    # Determine if a y axis was provided
    column_header_y_axis, sheet_index_y_axis =  None, None

    if 'column_header_y_axis' in keys:
        column_header_y_axis = event['column_header_y_axis']
    if 'sheet_index_y_axis' in keys:
        sheet_index_y_axis = event['sheet_index_y_axis']

    # See if we should include axes or not, by default we do not
    include_axes = False
    if 'include_axes' in keys:
        include_axes = event['include_axes']
   
    try:
        # When the x and y axis are specified in the API call, create a scatter plot 
        if column_header_y_axis is not None and sheet_index_y_axis is not None:
            # Get the y axis series
            sheet_index_y_axis = event['sheet_index_y_axis']
            y_series: pd.Series = wsc.dfs[sheet_index_y_axis][column_header_y_axis]

            # Create the scatter plot
            ax = sns.scatterplot(x=x_series, y=y_series)

            # Remove the x and y labels as we make them on the frontend
            # so that they are clickable
            if not include_axes:
                ax.set(xLabel=None, yLabel=None) 

            # Set the title of the graph
            ax.set_title(f'{column_header_x_axis} vs {column_header_y_axis}')

            # Remove the top and right border for better frontend display
            sns.despine(bottom=False, left=False) 

        # If only the x axis is specified in the API call, create a histogram or bar graph
        else:
            mito_type = get_mito_type(x_series)

            if mito_type == 'number_series':
                ax = x_series.plot.hist()
            else:
                if x_series.nunique() > 20:
                    # Send an empty response
                    send({
                        'event': 'api_response',
                        'id': event['id'],
                        'data': ''
                    })
                    # Then return so we only respond with an empty string
                    return
                ax = x_series.value_counts().plot.bar()

            ax.set_title(column_header_x_axis + ' frequencies')

        # Send the graph back to the frontend
        # From here: https://stackoverflow.com/questions/38061267/matplotlib-graphic-image-to-base64
        pic_IObytes = io.BytesIO()
        ax.figure.savefig(pic_IObytes, format='png', transparent=True, bbox_inches='tight')
        pic_IObytes.seek(0)
        pic_hash = base64.b64encode(pic_IObytes.read()).decode("utf-8").replace("\n", "")
        # Close all graphs so they don't each other write each
        plt.close('all')
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': pic_hash
        })
    except Exception as e:
        # As not being able to make a graph is a non-critical error that doesn't
        # result from user interaction, we don't want to throw an error if something
        # weird happens, so we just return nothing in this case
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': ''
        })
    
    # Turn interactive plotting back on, if it was on in the first place,
    # so we dont mess with users matplotlib code
    if was_interactive:
        plt.ion()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.

"""
A filter steps, which allows you to filter a column based on some
conditions and some values. 

NOTE: there are some workarounds in this step with old filter steps being deleted
and remade (like the one right) - simply because we don't have a way of editing
old steps otherwise!
"""
import functools
from mitosheet.utils import create_new_step, does_sheet_index_exist_within_step
from numbers import Number
import pandas as pd
from datetime import date

from mitosheet.errors import (
    EditError,
    make_execution_error,
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_filter_error
)

FILTER_COLUMN_DISPLAY_NAME = 'Filtered a Column'

FILTER_COLUMN_EVENT = 'filter_column_edit'
FILTER_COLUMN_STEP_TYPE = 'filter_column'

FILTER_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # column to filter from
    'operator', # 'Or' or 'And'
    # Filters are either a tripples of ({'type': <type>, 'condition': <condition>, 'value': <value>}),
    # or they are objects with {'filters': [list of above tripples], 'operator': 'Or' or 'And'}
    'filters', 
]


# CONSTANTS USED IN THE FILTER STEP ITSELF
STRING_TYPE = 'string'
NUMBER_TYPE = 'number'
DATETIME_TYPE = 'datetime'

SHARED_FILTER_CONDITIONS = [
    'empty',
    'not_empty'
]

STRING_FILTER_CONDITIONS = [
    'contains',
    'string_does_not_contain',
    'string_exactly'
]

NUMBER_FILTER_CONDITIONS = [
    'number_exactly',
    'greater',
    'greater_than_or_equal',
    'less',
    'less_than_or_equal'
]

DATETIME_FILTER_CONDITIONS = [
    'datetime_exactly',
    'datetime_greater',
    'datetime_greater_than_or_equal',
    'datetime_less',
    'datetime_less_than_or_equal',
]

def saturate_filter_event(curr_step, event):
    """
    Saturates the filter event with a `has_non_empty_filter` - which is useful
    for for logging
    """
    has_non_empty_filter = False
    for filter_or_group in event['filters']:
        if 'filters' in filter_or_group:
            # If it's a group
            if len(filter_or_group['filters']) > 0:
                has_non_empty_filter = True
        else:
            # If it's a single filter
            has_non_empty_filter = True

    event['has_non_empty_filter'] = has_non_empty_filter



def execute_filter_column(
        wsc,
        sheet_index,
        column_header,
        operator,
        filters
    ):
    """
    Filters an existing sheet with the given filters, which 
    each contain a condition and an optional value.
    """
    last_step = wsc.steps[-1]

    # if the sheet doesn't exist, throw an error
    if not does_sheet_index_exist_within_step(last_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the filtered column exists 
    missing_column = set([column_header]).difference(last_step['column_metatype'][sheet_index].keys())
    if len(missing_column) > 0: 
        raise make_no_column_error(missing_column)

    # if there is a filter already applied to the column, remove it
    _reset_filter(wsc, sheet_index, column_header)

    # if the condition is none, don't create a new step and delete the current filter if it exists
    if len(filters) == 0:
        # if there is a filter already applied to the column, remove it
        return

    last_step = wsc.steps[-1]


    # If no errors we create a new step for this filter
    new_step = create_new_step(last_step, FILTER_COLUMN_STEP_TYPE)
    # NOTE: we actually have to save these parameters in the step, as the filter step acts
    # differently than all other steps, tragically
    new_step['sheet_index'] = sheet_index
    new_step['column_header'] = column_header
    new_step['filters'] = filters
    new_step['operator'] = operator

    # Then we update the dataframe, first by executing on a fake dataframe
    try:
        # TODO: Remove this speculative execution, when it's safe to do so
        # (e.g. when we have proper step editing). 
        # For now, because filter is wacky, we leave it in here... 

        # make a copy of our data frame to test operate on 
        df_copy = new_step['dfs'][sheet_index].copy(deep=True)

        # We execute on the copy first to see if there will be errors
        _execute_filter(
            df_copy, 
            column_header,
            operator,
            filters
        )

    except EditError as e:
        # We propagate this error upwards
        raise e
    except Exception as e:
        print(e)
        # We raise a generic execution error in this case!
        raise make_execution_error()


    # if there was no error in execution on the copy, execute on real dataframes
    new_step['dfs'][sheet_index] = _execute_filter(
        new_step['dfs'][sheet_index], 
        column_header,
        operator,
        filters
    )

    # keep track of which columns are filtered
    new_step['column_filters'][sheet_index][column_header]['operator'] = operator
    new_step['column_filters'][sheet_index][column_header]['filters'] = filters

    wsc.steps.append(new_step)
    wsc.curr_step_idx = len(wsc.steps) - 1


def _reset_filter(wsc, sheet_index, column_header, ignore_last_step=False):
    """
    To reset the filter, we:

    1. Delete the filter step.
    2. Save the analysis.
    3. Delete _all_ the steps (ik, crazy).
    4. Replay the analysis.

    NOTE: we do this as a workaround because we want users to be able to edit
    filters, but there is currently no way to go back to a step and edit it. 

    We should _heavily_ re-evaluate this screwy-ness when we allow editing of steps, 
    and rolling back to steps.
    """
    # NOTE: we import here to avoid circular imports. This is an unfortunate
    # side effect of this entire workaround :( )
    from mitosheet.updates.replay_analysis import execute_replay_analysis_update
    from mitosheet.save_utils import write_analysis

    deleted_step_idx = None

    # find the step to delete
    for step_idx, step in enumerate(wsc.steps):
        if ignore_last_step and step_idx == wsc.curr_step_idx:
            continue

        if step['step_type'] == FILTER_COLUMN_STEP_TYPE:
            # If this is the step that added a filter to the column, mark it as the one to delete
            if step['sheet_index'] == sheet_index and step['column_header'] == column_header:
                deleted_step_idx = step_idx

    # if there are no step to delete, return
    if deleted_step_idx == None:
        return
    
    # Delete the filter step
    wsc.steps.pop(deleted_step_idx)

    # save the current analysis, now without the filter step
    write_analysis(wsc)

    # Delete _all the steps_ except the first
    wsc.steps = [wsc.steps[0]]

    # Rerun the analysis
    execute_replay_analysis_update(wsc, wsc.analysis_name, None, False)


def get_applied_filter(df, column_header, filter_):
    """
    Given a filter triple, returns the filter indexes for that
    actual dataframe
    """
    type_ = filter_['type']
    condition = filter_['condition']
    value = filter_['value']

    # First, we case on the shared conditions, to get them out of the way
    if condition in SHARED_FILTER_CONDITIONS:
        if condition == 'empty':
            return df[column_header].isna()
        elif condition == 'not_empty':
            return df[column_header].notnull()

    if type_ == STRING_TYPE:
        if condition not in STRING_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to string filter {condition}')

        # Check that the value is the valid
        if not isinstance(value, str):
            raise make_invalid_filter_error(value, STRING_TYPE)

        if condition == 'contains':
            return df[column_header].str.contains(value, na=False)
        if condition == 'string_does_not_contain':
            return ~df[column_header].str.contains(value, na=False)
        elif condition == 'string_exactly':
            return df[column_header] == value

    elif type_ == NUMBER_TYPE:
        if condition not in NUMBER_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to number filter {condition}')
        
        # Check that the value is the valid
        if not isinstance(value, Number):
            raise make_invalid_filter_error(value, NUMBER_TYPE)

        if condition == 'number_exactly':
            return df[column_header] == value
        elif condition == 'greater':
            return df[column_header] > value
        elif condition == 'greater_than_or_equal':
            return df[column_header] >= value
        elif condition == 'less':
            return df[column_header] < value
        elif condition == 'less_than_or_equal':
            return df[column_header] <= value

    elif type_ == DATETIME_TYPE:
        if condition not in DATETIME_FILTER_CONDITIONS:
            raise Exception(f'Invalid condition passed to datetime filter {condition}')

        # Check that we were given something that can be understood as a date
        try:
            timestamp = pd.to_datetime(value)
        except:
            # If we hit an error, because we restrict the input datetime, 
            # this is probably occuring because the user has only partially input the date, 
            # and so in this case, we just default it to the minimum possible timestamp for now!
            timestamp = date.min

        if condition == 'datetime_exactly':
            return df[column_header] == timestamp
        elif condition == 'datetime_greater':
            return df[column_header] > timestamp
        elif condition == 'datetime_greater_than_or_equal':
            return df[column_header] >= timestamp
        elif condition == 'datetime_less':
            return df[column_header] < timestamp
        elif condition == 'datetime_less_than_or_equal':
            return df[column_header] <= timestamp
    else:
        raise Exception(f'Invalid type passed in filter {type_}')

def combine_filters(operator, filters):

    def filter_reducer(filter_one, filter_two):
        # helper for combining filters based on the operations
        if operator == 'Or':
            return (filter_one) | (filter_two)
        elif operator == 'And':
            return (filter_one) & (filter_two)
        else:
            raise Exception(f'Operator {operator} is unsupported')

    # Combine all the filters into a single filter
    return functools.reduce(filter_reducer, filters)

def _execute_filter(
        df, 
        column_header,
        operator,
        filters
    ):
    """
    Executes a filter on the given column, filtering by removing any rows who
    don't meet the condition.
    """

    applied_filters = []

    for filter_or_group in filters:

        # If it's a group, then we build the filters for the group, combine them
        # and then add that to the applied filters
        if 'filters' in filter_or_group:
            group_filters = []
            for filter_ in filter_or_group['filters']:
                group_filters.append(
                    get_applied_filter(df, column_header, filter_)
                )

            applied_filters.append(
                combine_filters(filter_or_group['operator'], group_filters)
            )    

        # Otherwise, we just get that specific filter, and append it
        else:
            applied_filters.append(get_applied_filter(df, column_header, filter_or_group))    
    
    return df[combine_filters(operator, applied_filters)].reset_index(drop=True)


def get_filter_string(df_name, column_header, filter_):
    """
    Transpiles a specific filter to a fitler string, to be used
    in constructing the final transpiled code
    """
    condition = filter_['condition']
    value = filter_['value']

    FILTER_FORMAT_STRING_DICT = {
        # SHARED CONDITIONS
        'empty': '{df_name}.{column_header}.isna()',
        'not_empty': '{df_name}.{column_header}.notnull()',

        # NUMBERS
        'number_exactly': '{df_name}[\'{column_header}\'] == {value}',
        'greater': '{df_name}[\'{column_header}\'] > {value}',
        'greater_than_or_equal': '{df_name}[\'{column_header}\'] >= {value}',
        'less': '{df_name}[\'{column_header}\'] < {value}',
        'less_than_or_equal': '{df_name}[\'{column_header}\'] <= {value}',
        
        # STRINGS
        'contains': '{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
        'string_does_not_contain': '~{df_name}[\'{column_header}\'].str.contains(\'{value}\', na=False)',
        'string_exactly': '{df_name}[\'{column_header}\'] == \'{value}\'',

        # DATES
        'datetime_exactly': '{df_name}[\'{column_header}\'] == pd.to_datetime(\'{value}\')',
        'datetime_greater': '{df_name}[\'{column_header}\'] > pd.to_datetime(\'{value}\')',
        'datetime_greater_than_or_equal': '{df_name}[\'{column_header}\'] >= pd.to_datetime(\'{value}\')',
        'datetime_less': '{df_name}[\'{column_header}\'] < pd.to_datetime(\'{value}\')',
        'datetime_less_than_or_equal': '{df_name}[\'{column_header}\'] <= pd.to_datetime(\'{value}\')',            
    }

    return FILTER_FORMAT_STRING_DICT[condition].format(
        df_name=df_name,
        column_header=column_header,
        value=value
    )

def combine_filter_strings(operator, filter_strings, split_lines=False):
    """
    Combines the given filter strings with the passed operator, optionally 
    splitting the lines at 120 characters.
    
    NOTE: we choose to keep groups together for readibility, and so do not
    split the lines if we are combing a group.
    """
    if len(filter_strings) == 1:
        return filter_strings[0]
    else:
        # If there are multiple conditions, we combine them together, with the
        # given operator in the middle
        OPERATOR_SIGNS = {
            'Or': '|',
            'And': '&'
        }
        # Put parens around them
        filter_strings = [
            f'({fs})' for fs in filter_strings
        ]

        filter_string = ''
        current_line_length = 0
        for i, fs in enumerate(filter_strings):
            if i != 0:
                filter_string += f' {OPERATOR_SIGNS[operator]} '
            filter_string += fs
            # We keep track of how long the lines are, and if they go over 100 characters,
            # then we split them into a new line (not if this is the last one though)
            current_line_length += len(fs)
            if split_lines and current_line_length > 100 and i != len(filter_strings) - 1:
                filter_string += ' \\\n\t'
                current_line_length = 0

        return filter_string


def transpile_filter_column(
        step,
        sheet_index,
        column_header,
        operator,
        filters
    ):
    """
    Transpiles a filter step to Python code. 
    """
    df_name = step['df_names'][sheet_index]

    filter_strings = []
    for filter_or_group in filters:
        # If it is a group, we build the code for each filter, and then combine them at the end
        if 'filters' in filter_or_group:
            group_filter_strings = []
            for filter_ in filter_or_group['filters']:
                group_filter_strings.append(
                    get_filter_string(df_name, column_header, filter_)
                )

            filter_strings.append(
                # Note: we add parens around this, so it's grouped properly
                "(" + combine_filter_strings(filter_or_group['operator'], group_filter_strings) + ")"
            )
        else:
            filter_strings.append(
                get_filter_string(df_name, column_header, filter_or_group)
            )

    if len(filter_strings) == 0:
        return []
    elif len(filter_strings) == 1:
        return [
            f'{df_name} = {df_name}[{filter_strings[0]}]',
            f'{df_name} = {df_name}.reset_index(drop=True)'
        ]
    else:
        filter_string = combine_filter_strings(operator, filter_strings, split_lines=True)
        return [
            f'{df_name} = {df_name}[{filter_string}]',
            f'{df_name} = {df_name}.reset_index(drop=True)'
        ]


def describe_filter_column(
        sheet_index,
        column_header,
        operator,
        filters,
        df_names=None
    ):

    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Filtered {column_header} in {df_name}'
    return f'Filtered {column_header}'


FILTER_STEP = {
    'step_version': 1,
    'step_display_name': FILTER_COLUMN_DISPLAY_NAME,
    'event_type': FILTER_COLUMN_EVENT,
    'step_type': FILTER_COLUMN_STEP_TYPE,
    'params': FILTER_COLUMN_PARAMS,
    'saturate': saturate_filter_event,
    'execute': execute_filter_column,
    'transpile': transpile_filter_column,
    'describe': describe_filter_column
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A delete_column step, which allows you to delete a column
from a dataframe.
"""

from mitosheet.utils import create_new_step, does_sheet_index_exist_within_step
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_column_delete_error
)

DELETE_COLUMN_DISPLAY_NAME = 'Deleted a Column'

DELETE_COLUMN_EVENT = 'delete_column_edit'
DELETE_COLUMN_STEP_TYPE = 'delete_column'

DELETE_COLUMN_PARAMS = [
    'sheet_index', # the sheet to delete the column from
    'column_header', # the column to delete
]

def execute_delete_column_step(
        curr_step,
        sheet_index,
        column_header
    ):
    """
    Creates a new step that deletes a the column column_heder 
    from the sheet at sheet_index.
    """

    # if the sheet doesn't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    # Error if the column does not exist
    if column_header not in curr_step['column_metatype'][sheet_index]:
        raise make_no_column_error([column_header])
    
    # Error if there are any columns that currently rely on this column
    if len(curr_step['column_evaluation_graph'][sheet_index][column_header]) > 0:
        raise make_invalid_column_delete_error(
            column_header,
            list(curr_step['column_evaluation_graph'][sheet_index][column_header])
        )

    # Make a new step for the delete
    new_step = create_new_step(curr_step, DELETE_COLUMN_STEP_TYPE)
    
    # Actually drop the column
    df = new_step['dfs'][sheet_index]
    df.drop(column_header, axis=1, inplace=True)

    # And then update all the state variables removing this column from the state
    del new_step['column_metatype'][sheet_index][column_header]
    del new_step['column_type'][sheet_index][column_header]
    del new_step['column_spreadsheet_code'][sheet_index][column_header]
    del new_step['column_python_code'][sheet_index][column_header]
    del new_step['column_evaluation_graph'][sheet_index][column_header]
    # We also have to delete the places in the graph where this node is 
    for dependents in new_step['column_evaluation_graph'][sheet_index].values():
        if column_header in dependents:
            dependents.remove(column_header)
    
    return new_step


def transpile_delete_column_step(
        step,
        sheet_index,
        column_header
    ):
    df_name = step['df_names'][sheet_index]
    return [f'{df_name}.drop(\'{column_header}\', axis=1, inplace=True)']


def describe_delete_column_step(
        sheet_index,
        column_header,
        df_names=None
    ):

    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Deleted column {column_header} from {df_name}'
    return f'Deleted column {column_header}'


DELETE_COLUMN_STEP = {
    'step_version': 1,
    'step_display_name': DELETE_COLUMN_DISPLAY_NAME,
    'event_type': DELETE_COLUMN_EVENT,
    'step_type': DELETE_COLUMN_STEP_TYPE,
    'params': DELETE_COLUMN_PARAMS,
    'saturate': None,
    'execute': execute_delete_column_step,
    'transpile': transpile_delete_column_step,
    'describe': describe_delete_column_step
}